# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

from . import build, config, context, pack, test

__all__ = ["build", "config", "context", "pack", "test"]



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

import argparse
import inspect
import typing
from dataclasses import dataclass
from types import ModuleType
from typing import Annotated, Any, Dict, List, Optional, Tuple, Union, cast

from .. import commands
from .arg import Argument
from .config import Configs, FlowConfig, Runtime, default_compiler, platform


@dataclass
class SpecialArg:
    name: str
    ctor: callable


@dataclass
class EntryArg:
    name: str
    opt: bool
    help: str
    pos: bool
    names: List[str]
    nargs: Union[str, int, None]
    meta: Optional[str]
    action: Union[str, argparse.Action, None]
    default: Optional[Any]
    choices: Optional[List[str]] = None

    def visit(self, parser: argparse.ArgumentParser):
        kwargs = {}
        if self.help is not None:
            kwargs["help"] = self.help
        if self.nargs is not None:
            kwargs["nargs"] = self.nargs
        if self.meta is not None:
            kwargs["metavar"] = self.meta
        if self.default is not None:
            kwargs["default"] = self.default
        if self.action is not None:
            kwargs["action"] = self.action
        if self.choices is not None:
            kwargs["choices"] = self.choices

        names = (
            [self.name]
            if self.pos
            else self.names if len(self.names) > 0 else [f"--{self.name}"]
        )

        if self.pos:
            kwargs["nargs"] = "?" if self.opt else 1
        else:
            kwargs["dest"] = self.name
            kwargs["required"] = not self.opt

        parser.add_argument(*names, **kwargs)


@dataclass
class BuiltinEntry:
    name: str
    doc: str
    entry: callable
    args: List[EntryArg]
    additional: List[SpecialArg]

    def run(self, args: argparse.Namespace, cfg: FlowConfig):
        kwargs = {}
        for arg in self.args:
            kwargs[arg.name] = getattr(args, arg.name, None)

        for additional in self.additional:
            ctor = additional.ctor
            arg = ctor(args)
            kwargs[additional.name] = arg
            if ctor == Runtime:
                rt = cast(Runtime, arg)
                rt.steps = cfg.steps
                rt.aliases = cfg.aliases
                rt._cfg = cfg._cfg

        result = self.entry(**kwargs)
        return 0 if result is None else result

    @staticmethod
    def run_entry(args: argparse.Namespace, cfg: FlowConfig):
        builtin_entries = {entry.name for entry in command_list}
        aliases = cfg.aliases

        if args.command in builtin_entries:
            command = filter(
                lambda command: command.name == args.command, command_list
            ).__next__()
            return cast(BuiltinEntry, command).run(args, cfg)
        elif args.command in {alias.name for alias in aliases}:
            command = filter(
                lambda command: command.name == "run", command_list
            ).__next__()
            alias = filter(lambda alias: alias.name == args.command, aliases).__next__()
            args.steps.append(alias.steps)
            return cast(BuiltinEntry, command).run(args, cfg)

        print("known commands:")
        for command in command_list:
            print(f"   {command.name}: {command.doc}")
        for alias in aliases:
            print(f"   {alias.name}: {alias.doc}")
        return 1

    @staticmethod
    def visit_all(
        parser: argparse.ArgumentParser, cfg: FlowConfig
    ) -> Dict[str, List[str]]:
        shortcut_configs = BuiltinEntry.build_shortcuts(cfg)

        subparsers = parser.add_subparsers(
            dest="command", metavar="{command}", help="known command name, see below"
        )

        run: BuiltinEntry = None
        for entry in command_list:
            entry.visit(shortcut_configs, subparsers)
            if entry.name == "run":
                run = entry

        if run is not None and len(cfg.aliases) > 0:
            builtin_entries = {entry.name for entry in command_list}
            cfg.aliases = [
                alias for alias in cfg.aliases if alias.name not in builtin_entries
            ]
            for alias in cfg.aliases:
                run.visit(shortcut_configs, subparsers, alias=alias.name, doc=alias.doc)
        else:
            cfg.aliases = []

        return shortcut_configs

    @staticmethod
    def build_shortcuts(cfg: FlowConfig) -> Dict[str, List[str]]:
        shortcut_configs: Dict[str, List[str]] = {}
        args: List[Tuple[str, List[str]]] = []

        shortcuts = cfg.shortcuts
        for shortcut_name in sorted(shortcuts.keys()):
            shortcut = shortcuts[shortcut_name]
            config: List[str] = []
            for key in sorted(shortcut.keys()):
                value = shortcut[key]
                if isinstance(value, list):
                    for v in value:
                        config.append(f"{key}={_shortcut_value(v)}")
                else:
                    config.append(f"{key}={_shortcut_value(value)}")
            if len(config) > 0:
                args.append((shortcut_name, config))

        if len(args):
            os_prefix = f"os={platform}"
            compiler_prefix = f"compiler={default_compiler()}"

            for shortcut_name, config in args:
                config.insert(0, compiler_prefix)
                config.insert(0, os_prefix)
                shortcut_configs[shortcut_name] = config

        return shortcut_configs

    def visit(
        self,
        shortcut_configs: Dict[str, List[str]],
        subparsers,
        alias: Optional[str] = None,
        doc: Optional[str] = None,
    ):
        if not doc:
            doc = self.doc
        if not alias:
            alias = self.name

        parser: argparse.ArgumentParser = subparsers.add_parser(
            alias, help=doc, description=doc
        )

        parser.add_argument(
            "--dry-run",
            action="store_true",
            required=False,
            help="print steps and commands, do nothing",
        )

        parser.add_argument(
            "--silent",
            action="store_true",
            required=False,
            help="removes most of the output",
        )

        has_config = False
        for additional in self.additional:
            if additional.ctor == Configs:
                has_config = True
                break

        if has_config:
            parser.add_argument(
                "-D",
                dest="configs",
                metavar="config",
                nargs="*",
                action="append",
                default=[],
                help="run only build matching the config; "
                "each filter is a name of a matrix axis followed by comma-separated values to take; "
                f'if "os" is missing, it will default to additional "-D os={platform}"',
            )

            parser.add_argument(
                "--official",
                action="store_true",
                required=False,
                help="cut matrix to minimal set of builds",
            )

            if len(shortcut_configs):
                group = parser.add_mutually_exclusive_group()

                for shortcut_name in sorted(shortcut_configs.keys()):
                    config = shortcut_configs[shortcut_name]
                    group.add_argument(
                        f"--{shortcut_name}",
                        required=False,
                        action="store_true",
                        help=f'shortcut for "-D {" ".join(config)}"',
                    )

        for arg in self.args:
            arg.visit(parser)


def _shortcut_value(value) -> str:
    if isinstance(value, bool):
        return "ON" if value else "OFF"
    return str(value)


def _extract_arg(name: str, arg: Any):
    for ctor in [Configs, Runtime]:
        if arg is ctor:
            return SpecialArg(name, ctor)

    anno_type = typing.get_origin(arg)
    if anno_type is not Annotated:
        return None

    arg_dict = getattr(arg, "__dict__", {})
    origin = arg_dict.get("__origin__", None)
    (metadata,) = arg_dict.get("__metadata__", None)

    if origin is None or not isinstance(metadata, Argument):
        return None

    optional = typing.get_origin(origin) is Union and type(None) in typing.get_args(
        origin
    )
    return EntryArg(name, optional, **metadata.__dict__)


def _extract_args(entry: callable) -> List[Union[EntryArg, SpecialArg]]:
    entry_args = inspect.get_annotations(entry)
    nullable_args = (
        _extract_arg(name, entry_args[name]) for name in sorted(entry_args.keys())
    )
    args = filter(lambda item: item is not None, nullable_args)
    return list(args)


def _get_entry(modname: str, module: ModuleType):
    names = ["main", modname, f"command_{modname}"]
    for name, entry in inspect.getmembers(module):
        if not inspect.isfunction(entry) or name not in names:
            continue

        doc = inspect.getdoc(entry)
        args = _extract_args(entry)
        special_args = [entry for entry in args if isinstance(entry, SpecialArg)]
        entry_args = [entry for entry in args if isinstance(entry, EntryArg)]

        has_rt = False
        for additional in special_args:
            if additional.ctor == Runtime:
                has_rt = True
                break

        if not has_rt:
            continue

        return BuiltinEntry(modname, doc, entry, entry_args, special_args)


def _get_entries():
    submodules = inspect.getmembers(commands, inspect.ismodule)
    all_entries = map(lambda tup: _get_entry(*tup), submodules)
    valid_entries = filter(lambda entry: entry is not None, all_entries)
    return list(valid_entries)


command_list = _get_entries()



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

import os
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import List

from ..flow import matrix
from .config import Config, Runtime


class Step(ABC):
    @property
    @abstractmethod
    def name(self) -> str: ...

    @property
    def runs_after(self) -> List[str]:
        return []

    @property
    def runs_before(self) -> List[str]:
        return []

    def platform_dependencies(self) -> List[str]:
        return []

    def is_active(self, config: Config, rt: Runtime) -> bool:
        return True

    def directories_to_remove(self, config: Config) -> List[str]:
        return []

    @abstractmethod
    def run(self, config: Config, rt: Runtime) -> int: ...


class SerialStep(Step):
    children: List[Step] = []

    @property
    def runs_after(self):
        return matrix.flatten([child.runs_after for child in self.children])

    def platform_dependencies(self) -> List[str]:
        return matrix.flatten(
            [child.platform_dependencies() for child in self.children]
        )

    def is_active(self, config: Config, rt: Runtime) -> bool:
        for child in self.children:
            if not child.is_active(config, rt):
                return False
        return True

    def directories_to_remove(self, config: Config) -> List[str]:
        return matrix.flatten(
            [child.directories_to_remove(config) for child in self.children]
        )

    def run(self, config: Config, rt: Runtime) -> int:
        for child in self.children:
            result = child.run(config, rt)
            if result:
                return result
        return 0


@dataclass
class Statement:
    rule: "Rule"
    outputs: List[str]
    inputs: List[str]
    implicit_deps: List[str] = field(default_factory=list)

    def run(self, rt: Runtime):
        command = self.rule.command(self)
        if len(command) == 0:
            return self._run_directly(rt)

        if rt.dry_run:
            rt.print(*command)
            return 0

        if not self._needed():
            return False

        return rt.cmd(*command)

    def _run_directly(self, rt: Runtime):
        if rt.dry_run:
            copy = Runtime(rt)
            copy.dry_run = True
            return self.rule.run(self, copy)

        if not self._needed():
            return False

        return self.rule.run(self, rt)

    def _needed(self):
        out_mtime = None
        for out in self.outputs:
            try:
                mtime = os.path.getmtime(out)
                out_mtime = mtime if out_mtime is None else min(mtime, out_mtime)
            except FileNotFoundError:
                pass
        if out_mtime is None:
            return True

        dep_mtime = 0
        for deps in [self.inputs, self.implicit_deps]:
            for dep in deps:
                try:
                    mtime = os.path.getmtime(dep)
                    dep_mtime = max(dep_mtime, mtime)
                except FileNotFoundError:
                    pass
        return dep_mtime > out_mtime


class Rule(ABC):
    @abstractmethod
    def command(self, statement: Statement) -> List[str]: ...

    def run(self, statement: Statement, rt: Runtime):
        return 1

    @classmethod
    def statement(
        cls, outputs: List[str], inputs: List[str], implicit_deps: List[str] = []
    ):
        return Statement(cls(), outputs, inputs, implicit_deps)


@dataclass(init=False)
class Makefile:
    statements: List[Statement]

    @dataclass
    class Sorted:
        outputs: List[str]
        deps: List[str]
        ref: Statement

    def __init__(self, statements: List[Statement]):
        unsorted = [
            Makefile.Sorted(
                outputs=[*st.outputs], deps=[*st.inputs, *st.implicit_deps], ref=st
            )
            for st in statements
        ]
        sorted: List[Statement] = []

        for st in unsorted:
            copy = [*st.deps]
            for dep in copy:
                found = False
                for node in unsorted:
                    if dep in node.outputs:
                        found = True
                        break
                if not found:
                    st.deps.remove(dep)

        while len(unsorted):
            next = [st for st in unsorted if len(st.deps) == 0]
            unsorted = [st for st in unsorted if len(st.deps) > 0]

            sorted.extend(st.ref for st in next)
            for st in next:
                for output in st.outputs:
                    for left in unsorted:
                        if output in left.deps:
                            left.deps.remove(output)

        self.statements = sorted

    def run(self, rt: Runtime):
        counter = 0
        for statement in self.statements:
            result = statement.run(rt)
            if isinstance(result, bool):
                if not result:
                    counter += 1
            if result:
                return result
        if counter == len(self.statements):
            print("-- Nothing to do")
        return 0


__steps: List[Step] = []


def register_step(step: Step):
    global __steps

    name = step.name
    if name in [step.name for step in __steps]:
        raise NameError(f"Step {name} already registered")

    __steps.append(step)



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

import argparse
import os
import sys
from typing import List, cast

from .. import __version__
from .commands import BuiltinEntry
from .config import FlowConfig
from .steps import clean_aliases, load_steps


def _change_dir():
    root = argparse.ArgumentParser(
        prog="cxx-flow",
        usage="cxx-flow [-h] [--version] [-C [dir]] {command} ...",
        add_help=False,
    )
    root.add_argument("-C", dest="cd", nargs="?")

    args, _ = root.parse_known_args()
    if args.cd:
        os.chdir(args.cd)


def __main():
    _change_dir()

    flow_cfg = FlowConfig()
    valid_steps = load_steps(flow_cfg)
    clean_aliases(flow_cfg, valid_steps)

    root = argparse.ArgumentParser(
        prog="cxx-flow", description="C++ project maintenance, automated"
    )
    root.add_argument(
        "-v", "--version", action="version", version=f"%(prog)s version {__version__}"
    )
    root.add_argument(
        "-C",
        dest="cd",
        metavar="dir",
        nargs="?",
        help="runs as if cxx-flow was started in <dir> instead of the current working directory",
    )

    shortcut_configs = BuiltinEntry.visit_all(root, flow_cfg)
    args = root.parse_args()

    args_kwargs = dict(args._get_kwargs())
    for key in shortcut_configs:
        try:
            if not args_kwargs[key]:
                continue
            cast(List[List[str]], args.configs).append(shortcut_configs[key])
            break
        except KeyError:
            continue

    sys.exit(BuiltinEntry.run_entry(args, flow_cfg))


def main():
    try:
        __main()
    except KeyboardInterrupt:
        sys.exit(1)



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

import os
from pprint import pprint
from typing import Annotated, Dict, List, Optional, Set, cast

from ..flow import commands, matrix, step
from ..flow.arg import FlagArgument
from ..flow.config import Runtime


def command_list(
    builtin: Annotated[Optional[str], FlagArgument(help="show all builtin commands")],
    alias: Annotated[Optional[str], FlagArgument(help="show all alias commands")],
    steps: Annotated[Optional[str], FlagArgument(help="show all run steps")],
    configs: Annotated[Optional[str], FlagArgument(help="show all known matrix keys")],
    all: Annotated[
        Optional[str], FlagArgument(help="show builtins, aliases, steps and configs")
    ],
    pipe: Annotated[
        Optional[str], FlagArgument(help="do not show additional information")
    ],
    rt: Runtime,
):
    """List all the commands and/or steps for cxx-flow"""

    printed_something = False
    bold = "\033[96m" if rt.use_color else ""
    reset = "\033[0m" if rt.use_color else ""

    if all:
        builtin = True
        alias = True
        steps = True
        configs = True

    if builtin:
        builtin_entries = list(
            sorted((entry.name, entry.doc) for entry in commands.command_list)
        )
        if not pipe and len(builtin_entries) > 0:
            print("Builtin commands")
            print("----------------")

        for entry_name, entry_doc in builtin_entries:
            if pipe:
                print(entry_name)
                continue

            name = f"{bold}{entry_name}{reset}"
            if entry_doc:
                print(f"- {name}: {entry_doc}")
            else:
                print(f"- {name}")

        printed_something = True

    if alias:
        aliases = rt.aliases

        if not pipe and len(aliases) > 0:
            if printed_something:
                print()

            print("Known aliases")
            print("-------------")

        for alias in aliases:
            if pipe:
                print(alias.name)
                continue

            name = f"{bold}{alias.name}{reset}"
            print(f"- {name}: {", ".join(alias.steps)}")

        printed_something = True

    if steps:
        rt_steps = cast(List[step.Step], rt.steps)

        if not pipe and len(rt_steps) > 0:
            if printed_something:
                print()

            print("Run steps")
            print("---------")

        some_unused = False
        aliased_steps = set(matrix.flatten([alias.steps for alias in rt.aliases]))

        for rt_step in rt_steps:
            if pipe:
                print(rt_step.name)
                continue

            step_used = rt_step.name in aliased_steps
            if not step_used:
                some_unused = True

            name = f"{bold}{rt_step.name}{reset}"
            if step_used:
                print(f"- {name}")
            else:
                print(f"- {name}*")

        if some_unused:
            print(
                f"*step can only be run by explicitly calling through {bold}run{reset}."
            )

        printed_something = True

    if configs:
        m, keys = _load_flow_data(rt)
        if pipe:
            for key in keys:
                print(key)
        else:
            if len(keys) > 0:
                if printed_something:
                    print()

                print("Matrix keys")
                print("-----------")

            values: Dict[str, Set[str]] = {}

            for config in m:
                for key in keys:
                    value = config[key]
                    if isinstance(value, bool):
                        value = "ON" if value else "OFF"
                    else:
                        value = str(value)
                    try:
                        values[key].add(value)
                    except KeyError:
                        values[key] = {value}

            empty = set()
            for key in keys:
                value = ", ".join(values.get(key, empty))
                name = f"{bold}{key}{reset}"
                if value:
                    print(f"- {name}: {value}")
                else:
                    print(f"- {name}")

        printed_something = True

    if not printed_something and not pipe:
        print(f"Use {bold}--help{reset} to see, which listings are available")


def _load_flow_data(rt: "Runtime"):
    paths = [os.path.join(".flow", "matrix.json")]
    m, keys = matrix.load_matrix(*paths)

    if rt.no_coverage:
        for conf in m:
            if "coverage" in conf:
                del conf["coverage"]

    return m, keys



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

import json
import os
import sys
from typing import Annotated, Optional

from ..flow import config, ctx, dependency, init
from ..flow.arg import Argument, FlagArgument
from ..flow.interact import prompt
from ..flow.layer import copy_license, gather_package_layers


def command_init(
    path: Annotated[
        Optional[str],
        Argument(
            help="location of initialized project; defaults to current directory",
            pos=True,
            default=".",
        ),
    ],
    non_interactive: Annotated[
        Optional[bool],
        FlagArgument(help="selects all the default answers", names=["-y", "--yes"]),
    ],
    save_context: Annotated[
        Optional[bool],
        FlagArgument(help="save the mustache context json", names=["--ctx"]),
    ],
    rt: config.Runtime,
):
    """Initializes a new project"""

    if path is not None:
        os.makedirs(path, exist_ok=True)
        os.chdir(path)

    errors = dependency.verify(dependency.gather(init.__steps))
    if len(errors) > 0:
        if not rt.silent:
            for error in errors:
                print(f"cxx-flow: {error}", file=sys.stderr)
        return 1

    context = ctx.fixup(ctx.all_default() if non_interactive else prompt())
    if not non_interactive and not rt.silent:
        print()

    if save_context:
        with open(".context.json", "w", encoding="UTF-8") as jsonf:
            json.dump(context, jsonf, ensure_ascii=False, indent=4)

    copy_license(rt, context)
    if not rt.silent:
        print()

    layers = gather_package_layers(ctx.package_root, context)
    for layer in layers:
        layer.run(rt, context)

    if save_context:
        with open(".gitignore", "ab") as ignoref:
            ignoref.write("\n/.context.json\n".encode("UTF-8"))

    for step in init.__steps:
        step.postprocess(rt, context)



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)


import json
import os
import sys
from typing import Dict, List, Tuple, TypeVar

T = TypeVar("T")


def find_compiler(
    compiler: str, config_names: Dict[str, List[str]]
) -> Tuple[str, List[str]]:
    dirname = os.path.dirname(compiler)
    filename = os.path.basename(compiler)
    if sys.platform == "win32":
        filename = os.path.splitext(filename)[0]
    chunks = filename.split("-", 1)
    if len(chunks) == 1:
        version = None
    else:
        version = chunks[1]
    filename = chunks[0].lower()

    try:
        compiler_names = config_names[filename]
    except:
        compiler_names = [filename]

    compilers = [
        os.path.join(dirname, name if version is None else f"{name}-{version}")
        for name in compiler_names
    ]

    if filename == "stdclang":
        filename = "clang"

    return filename, compilers


def flatten(array: List[List[T]]) -> List[T]:
    return [item for sublist in array for item in sublist]


def matches(tested: dict, test: dict):
    for key, value in test.items():
        val = tested.get(key)
        if val != value:
            return False
    return True


def matches_any(tested: dict, tests: List[dict]):
    for test in tests:
        if matches(tested, test):
            return True
    return False


def cartesian(input: Dict[str, list]) -> List[dict]:
    product = [{}]

    for key, values in input.items():
        next_level = []
        for value in values:
            for obj in product:
                next_level.append({**obj, key: value})
        product = next_level

    return product


def _split_keys(includes: List[dict], keys: List[str]) -> List[Tuple[dict, dict]]:
    result = []
    for include in includes:
        expand_key = {}
        expand_value = {}
        for key, value in include.items():
            if key in keys:
                expand_key[key] = value
            else:
                expand_value[key] = value
        result.append((expand_key, expand_value))
    return result


def load_matrix(*json_paths: str) -> Tuple[List[dict], List[str]]:
    setups: List[dict] = []
    for json_path in json_paths:
        try:
            with open(json_path, encoding="UTF-8") as f:
                setups.append(json.load(f))
        except FileNotFoundError:
            pass

    if len(setups) == 0:
        return [], []

    setup = setups[0]
    for additional in setups[1:]:
        src_matrix = setup.get("matrix", {})
        src_exclude = setup.get("exclude", [])
        src_include = setup.get("include", [])

        for key, value in additional.get("matrix", {}).items():
            old = src_matrix.get(key)
            if isinstance(old, list) and isinstance(value, list):
                old.extend(value)
            elif isinstance(old, list):
                old.append(value)
            else:
                src_matrix[key] = value
        src_exclude.extend(additional.get("exclude", []))
        src_include.extend(additional.get("include", []))

    raw = setup.get("matrix", {})
    keys = list(raw.keys())
    full = cartesian(raw)

    includes = _split_keys(setup.get("include", []), keys)
    for obj in full:
        for include_key, include_value in includes:
            if not matches(obj, include_key):
                continue
            for key, value in include_value.items():
                obj[key] = value

    excludes = setup.get("exclude", [])
    matrix = [obj for obj in full if not matches_any(obj, excludes)]

    return matrix, keys



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

import json
import os
import sys
from typing import Annotated, Optional

from ..flow.arg import FlagArgument
from ..flow.config import Configs, Runtime


def command_github(
    matrix: Annotated[Optional[bool], FlagArgument(help="print matrix json")],
    configs: Configs,
    rt: Runtime,
):
    """Supplies data for github actions"""

    if matrix:
        usable = [usable.items for usable in configs.usable]
        for config in usable:
            if "--orig-compiler" in config:
                orig_compiler = config["--orig-compiler"]
                del config["--orig-compiler"]
                config["compiler"] = orig_compiler
        if "GITHUB_ACTIONS" in os.environ:
            var = json.dumps({"include": usable})
            print(f"matrix={var}")
        else:
            json.dump(usable, sys.stdout)
        return

    print("cxx-flow github", matrix, configs)



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

import datetime
import inspect
import os
from dataclasses import dataclass
from pprint import pprint
from typing import Callable, Dict, Iterable, List, Optional, Union

from . import cmd

StrOrBool = Union[str, bool]
Values = Union[StrOrBool, List[str]]
SettingsType = Dict[str, StrOrBool]


@dataclass
class Setting:
    json_key: str
    prompt: str = ""
    value: Union[Values, Callable[[], Values]] = ""
    fix: Optional[str] = None
    force_fix: bool = False

    def calc_value(self, previous: SettingsType):
        if isinstance(self.value, Callable):
            kwargs = {}

            params = inspect.signature(self.value).parameters
            if "settings" in params:
                kwargs["settings"] = previous

            return self.value(**kwargs)

        return self.value


package_root = os.path.dirname(os.path.dirname(__file__))
template_dir = "template"


_fileext = {".cc": ".hh", ".cxx": ".hxx", ".cpp": ".hpp"}


def _git_config(name: str):
    def wrap():
        proc = cmd.run("git", "config", name, capture_output=True)
        if proc is None or proc.returncode != 0:
            return None
        return proc.stdout.strip()

    return wrap


def move_to_front(preferred: str, values: Iterable[str]):
    result: List[str] = []

    has_preferred = False
    for value in values:
        if value == preferred:
            has_preferred = True
        else:
            result.append(value)

    if has_preferred:
        result.insert(0, preferred)

    return result


def _list_ext():
    return move_to_front(".cpp", sorted(_fileext.keys()))


def _as_mustache(basename: str):
    name, ext = os.path.splitext(basename)
    return ext == ".mustache" and name or None


def _enum_licenses():
    root = os.path.abspath(os.path.join(package_root, template_dir, "licenses"))
    for _, dirnames, filenames in os.walk(root):
        dirnames[:] = []
        iter = filter(lambda x: x is not None, map(_as_mustache, filenames))
        return move_to_front("MIT", iter)
    return []


def _filter(
    src: Callable[[SettingsType], StrOrBool], flt: Callable[[StrOrBool], StrOrBool]
):
    def impl(settings: SettingsType):
        return flt(src(settings))

    return impl


def _get_nothing(_: SettingsType) -> StrOrBool:
    return ""


def _map(internal_key: str):
    def impl(key: str):
        mapped = internals.get(internal_key)
        if not isinstance(mapped, dict):
            return None
        return mapped.get(key)

    return impl


def _get_key(key: str):
    def impl(settings: SettingsType):
        return settings.get(key, "")

    return impl


_filters: Dict[str, Callable[[StrOrBool], StrOrBool]] = {
    "safe": lambda value: value.replace("-", "_"),
    "upper": lambda value: value.upper(),
    "lower": lambda value: value.lower(),
    "header": lambda cxx_ext: _fileext.get(cxx_ext, ".hpp"),
}


def _build_fixup(settings: SettingsType, fixup: str):
    fixup_refs = fixup.split("{")
    result = fixup_refs.pop(0)
    for ref_expr in fixup_refs:
        ref_text, verbose = ref_expr.split("}")
        filter_path = ref_text.split("$")

        ref = filter_path.pop(0)
        code = _get_nothing if ref == "" else _get_key(ref)

        for filter_name in filter_path:
            flt = _filters.get(filter_name)
            if flt is None:
                if filter_name.startswith("map:"):
                    internal_key = filter_name[4:]
                    code = _filter(code, _map(internal_key))
                continue
            code = _filter(code, flt)

        value = code(settings)

        if result:
            result += value
        else:
            result = value

        if verbose:
            result += verbose

    return result


def _fixup(settings: SettingsType, key: str, fixup: str, force=False):
    value = settings.get(key, "")
    if value != "" and not force:
        return

    value = _build_fixup(settings, fixup)
    settings[key] = value


def _fixed(fixup: str):
    def wrap(settings: SettingsType):
        return _build_fixup(settings, fixup)

    return wrap


def _get_default(setting: Setting, settings: SettingsType):
    value = setting.calc_value(settings)
    if isinstance(value, list):
        return value[0]
    return value


def all_default():
    settings: SettingsType = {}

    for setting in defaults:
        value = _get_default(setting, settings)
        settings[setting.json_key] = value

    for setting in switches:
        value = _get_default(setting, settings)
        settings[setting.json_key] = value

    return settings


def fixup(settings: SettingsType):
    for setting in hidden:
        value = _get_default(setting, settings)
        if isinstance(value, bool) or value != "":
            settings[setting.json_key] = value

    for coll in [defaults, hidden]:
        for setting in coll:
            _fixup(settings, setting.json_key, setting.fix or "", setting.force_fix)
    del settings["EXT"]

    result = {}
    for key in settings:
        path = key.split(".")
        ctx = result
        for step in path[:-1]:
            if step not in ctx or not isinstance(ctx[step], dict):
                ctx[step] = {}
            ctx = ctx[step]
        ctx[path[-1]] = settings[key]
    return result


def register_init_setting(*setting: Setting, is_hidden=False):
    (hidden if is_hidden else defaults).extend(setting)


def register_switch(key: str, prompt: str, enabled: bool):
    switches.append(Setting(key, prompt, value=enabled))


def register_internal(key: str, value: any):
    internals[key] = value


def get_internal(key: str, value: any = None):
    return internals.get(key, value)


internals = {}

switches: List[Setting] = []

defaults: List[Setting] = [
    Setting(
        "PROJECT.NAME",
        "Project name",
        lambda: os.path.basename(os.path.abspath("")) or "?project-name?",
    ),
    Setting("PROJECT.DESCRIPTION", "Project description"),
    Setting(
        "PROJECT.EMAIL",
        "Valid email, e.g. for CODE_OF_CONDUCT",
        _git_config("user.email"),
    ),
    Setting(
        "COPY.YEAR",
        "Year for copyright notices",
        str(datetime.date.today().year),
    ),
    Setting(
        "COPY.HOLDER",
        "Holder of the copyright",
        _git_config("user.name"),
    ),
    Setting("COPY.LICENSE", "License", _enum_licenses),
    Setting(
        "INCLUDE_PREFIX",
        'Prefix for includes (as in #include "{PREFIX}/version.hpp")',
        _fixed("{PROJECT.NAME}"),
    ),
    Setting(
        "NAME_PREFIX",
        "CMake variable name prefix",
        _fixed("{PROJECT.NAME$safe$upper}"),
    ),
    Setting(
        "NAMESPACE",
        "C++ namespace for the project",
        _fixed("{PROJECT.NAME$safe}"),
    ),
    Setting("EXT", "Extension for code files", _list_ext),
    Setting("SRCDIR", "Directory for code files", "src"),
    Setting(
        "INCLUDEDIR",
        "Directory for include files",
        "include",
        "{INCLUDEDIR}/{INCLUDE_PREFIX}",
        force_fix=True,
    ),
]

hidden: List[Setting] = [
    Setting("EXT.cxx", fix="{EXT}"),
    Setting("EXT.hxx", fix="{EXT.cxx$header}"),
]



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

import platform
import shlex


def _os_release_fallback():
    result = {}
    with open("/etc/os-release", encoding="UTF-8") as os_rel:
        for line in os_rel:
            line = line.strip()
            if line[:1] == "#":
                line = ""
            if line == "":
                continue
            name, value = line.split("=", 1)
            result[name] = " ".join(shlex.split(value))
    return result


def uname():
    _platform = platform.uname()
    platform_name = _platform.system.lower()
    platform_version = _platform.version
    platform_arch = _platform.machine.lower()

    if platform_arch == "amd64":
        platform_arch = "x86_64"

    system_nt = platform_name.split("_nt-", 1)

    if len(system_nt) > 1:
        platform_name = system_nt[0]
        platform_version = None
    elif platform_name == "windows":
        platform_version = None
    elif platform_name == "linux":
        try:
            os_release = platform.freedesktop_os_release()
        except AttributeError:
            os_release = _os_release_fallback()

        platform_id = os_release.get("ID", os_release.get("NAME"))
        version_id = os_release.get("VERSION_ID", platform_version)
        if platform_id is not None:
            if platform_id[:9] == "opensuse-":
                platform_id = "opensuse"
            platform_name = platform_id.lower()
            if version_id != "" and version_id[0] in "0123456789":
                platform_version = ".".join(version_id.split(".", 2)[:2])

    return (
        platform_name,
        platform_version,
        platform_arch,
    )



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

import base64
import json
import os
import platform
import struct
import subprocess
import sys
import winreg
from typing import Iterable, List, NamedTuple, Optional, Tuple

ENV_KEY = "SIGN_TOKEN"

Version = Tuple[int, int, int]

machine = {"ARM64": "arm64", "AMD64": "x64", "X86": "x86"}.get(
    platform.machine(), "x86"
)


def find_sign_tool() -> Optional[str]:
    with winreg.OpenKeyEx(
        winreg.HKEY_LOCAL_MACHINE, r"SOFTWARE\Microsoft\Windows Kits\Installed Roots"
    ) as kits:
        try:
            kits_root = winreg.QueryValueEx(kits, "KitsRoot10")[0]
        except FileNotFoundError:
            return None

        versions: List[Tuple[Version, str]] = []
        try:
            index = 0
            while True:
                ver_str = winreg.EnumKey(kits, index)
                ver = tuple(int(chunk) for chunk in ver_str.split("."))
                index += 1
                versions.append((ver, ver_str))
        except OSError:
            pass
    versions.sort()
    versions.reverse()
    for _, version in versions:
        # C:\Program Files (x86)\Windows Kits\10\bin\10.0.22621.0\x64\signtool.exe
        sign_tool = os.path.join(kits_root, "bin", version, machine, "signtool.exe")
        if os.path.isfile(sign_tool):
            return sign_tool
    return None


def get_key_():
    env = os.environ.get(ENV_KEY)
    if env is not None:
        return env
    filename = os.path.join(os.path.expanduser("~"), "signature.key")
    if os.path.isfile(filename):
        with open(filename, encoding="UTF-8") as file:
            return file.read().strip()
    filename = os.path.join(".", "signature.key")
    if os.path.isfile(filename):
        with open(filename, encoding="UTF-8") as file:
            return file.read().strip()

    return None


class Key(NamedTuple):
    token: str
    secret: bytes


def get_key() -> Optional[Key]:
    key = get_key_()
    if key is None:
        return None
    obj = json.loads(key)
    token = obj.get("token")
    secret = obj.get("secret")
    return Key(
        base64.b64decode(token).decode("UTF-8") if token is not None else None,
        base64.b64decode(secret) if secret is not None else None,
    )


def is_active(os_name: str):
    if os_name != "windows":
        return False
    key = get_key()
    return (
        find_sign_tool() is not None
        and key is not None
        and key.token is not None
        and key.secret is not None
    )


_IMAGE_DOS_HEADER = "HHHHHHHHHHHHHH8sHH20sI"
_IMAGE_NT_HEADERS_Signature = "H"
_IMAGE_DOS_HEADER_size = struct.calcsize(_IMAGE_DOS_HEADER)
_IMAGE_NT_HEADERS_Signature_size = struct.calcsize(_IMAGE_NT_HEADERS_Signature)
_MZ = 23117
_PE = 17744


def is_pe_exec(path: str):
    with open(path, "rb") as exe:
        mz_header = exe.read(_IMAGE_DOS_HEADER_size)
        dos_header = struct.unpack(_IMAGE_DOS_HEADER, mz_header)
        if dos_header[0] != _MZ:
            return False

        PE_offset = dos_header[-1]
        if PE_offset < _IMAGE_DOS_HEADER_size:
            return False

        if PE_offset > _IMAGE_DOS_HEADER_size:
            exe.read(PE_offset - _IMAGE_DOS_HEADER_size)

        pe_header = exe.read(_IMAGE_NT_HEADERS_Signature_size)
        signature = struct.unpack(_IMAGE_NT_HEADERS_Signature, pe_header)[0]
        return signature == _PE


def sign(files: Iterable[str]):
    key = get_key()

    if key is None or key.token is None or key.secret is None:
        print("cxx-flow: sign: the key is missing", file=sys.stderr)
        return 1

    sign_tool = find_sign_tool()
    if sign_tool is None:
        print("cxx-flow: sign: signtool.exe not found", file=sys.stderr)
        sys.exit(0)

    with open("temp.pfx", "wb") as pfx:
        pfx.write(key.secret)

    args = [
        sign_tool,
        "sign",
        "/f",
        "temp.pfx",
        "/p",
        key.token,
        "/tr",
        "http://timestamp.digicert.com",
        "/fd",
        "sha256",
        "/td",
        "sha256",
        *files,
    ]

    result = 1
    try:
        result = subprocess.run(args, shell=False).returncode
    finally:
        os.remove("temp.pfx")

    return result



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

import os

from cxx_flow.flow.config import Config, Runtime
from cxx_flow.flow.step import Step, register_step

from .__version__ import CMAKE_VERSION


class PackStep(Step):
    name = "Pack"
    runs_after = ["Build"]

    def platform_dependencies(self):
        return [f"cmake>={CMAKE_VERSION}", f"cpack>={CMAKE_VERSION}"]

    def is_active(self, config: Config, rt: Runtime) -> int:
        return (
            os.path.isfile("CMakeLists.txt")
            and os.path.isfile("CMakePresets.json")
            and len(config.items.get("cpack_generator", [])) > 0
        )

    def run(self, config: Config, rt: Runtime) -> int:
        generators = ";".join(config.items.get("cpack_generator", []))
        return rt.cmd("cpack", "--preset", config.preset, "-G", generators)


register_step(PackStep())



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

from . import store_both, store_packages, store_test

__all__ = ["store_both", "store_test", "store_packages"]
